from bs4 import PageElement, BeautifulSoup

from .options import Options


def make_cover(soup: PageElement, options: Options):
    """ Generate a cover pages.

    Arguments:
        soup {BeautifulSoup} -- target element.
        options {Options} -- the project options.
    """

    if options.cover:
        _make_cover(soup, options)

    if options.back_cover:
        _make_back_cover(soup, options)


def _make_cover(soup: PageElement, options: Options):
    try:
        keywords = options.template.keywords
        template = options.template.select(['cover', 'default_cover'])

        options.logger.info(f'Generate a cover page with "{template.name}".')
        soup_template = BeautifulSoup(template.render(keywords), 'html.parser')

        soup.body.insert(0, soup_template)
    except Exception as e:
        options.logger.error('Failed to generate the cover page: %s', e)


def _make_back_cover(soup: PageElement, options: Options):
    try:
        keywords = options.template.keywords
        template = options.template.select(
            ['back_cover', 'default_back_cover'])

        options.logger.info(
            f'Generate a back cover page with "{template.name}".')
        soup_template = BeautifulSoup(template.render(keywords), 'html.parser')

        soup.body.append(soup_template)
    except Exception as e:
        options.logger.error('Failed to generate the back cover page: %s', e)



import base64
import io
from enum import Enum

import qrcode
import qrcode.image.svg

from . import _FilterBase

#
# @see https://pypi.org/project/qrcode/
#


class _Format(Enum):
    SVG = 1
    PNG = 2


def _parse_format(format: str) -> _Format:
    if format.upper() == 'PNG':
        return _Format.PNG
    return _Format.SVG


def _parse_error_correction(error_correction: str):
    if error_correction:
        lv = error_correction.upper()
        if lv.startswith('Q'):
            return qrcode.constants.ERROR_CORRECT_Q
        elif lv.startswith('H'):
            return qrcode.constants.ERROR_CORRECT_H
        elif lv.startswith('L'):
            return qrcode.constants.ERROR_CORRECT_L
    return qrcode.constants.ERROR_CORRECT_M


def _image_factory(fmt: _Format):
    if fmt == _Format.PNG:
        return None
    return qrcode.image.svg.SvgPathImage


def _save_kind(fmt: _Format):
    if fmt == _Format.PNG:
        return 'PNG'
    return 'SVG'


def _content_type(fmt: _Format) -> str:
    if fmt == _Format.PNG:
        return 'image/png'
    return 'image/svg+xml;charset=utf-8'


class QRCode(_FilterBase):
    """ Generate QRCode for given value and returns embedded image data. """

    def __call__(self, content: str, format='SVG',
                 version=None,
                 error_correction=None,
                 box_size=10, border=4,
                 mask_pattern=None,
                 optimize=20):

        fmt = _parse_format(format)

        maker = qrcode.QRCode(
            version=version,
            error_correction=_parse_error_correction(error_correction),
            box_size=box_size,
            border=border,
            image_factory=_image_factory(fmt),
            mask_pattern=mask_pattern)
        maker.add_data(content, optimize=optimize)
        img = maker.make_image()

        with io.BytesIO() as stream:
            img.save(stream, kind=_save_kind(fmt))
            data = base64.b64encode(stream.getvalue())
            return f'data:{_content_type(fmt)};base64,' + data.decode("ascii")



import os
from urllib.parse import urlparse

from . import _FilterBase


class URLFilter(_FilterBase):
    """ Finds a matching filename in some directories and returns its URL. """

    def __call__(self, pathname: str) -> str:
        if not pathname:
            return ''

        # Check for URL(eg. 'https://...')
        target_url = urlparse(pathname)
        if target_url.scheme or target_url.netloc:
            return pathname

        # Search image file in below directories:
        dirs = [
            self.options.custom_template_path,
            getattr(self.config['theme'], 'custom_dir', None),
            self.config['docs_dir'],
            '.'
        ]

        for d in dirs:
            if not d:
                continue
            path = os.path.abspath(os.path.join(d, pathname))
            if os.path.isfile(path):
                return 'file://' + path

        # not found?
        return pathname



import html
import os

import sass

from ..options import Options


def _css_escape(text: str) -> str:
    """ @see https://developer.mozilla.org/en-US/docs/Web/CSS/string """

    if not text:
        return ''

    text = html.unescape(text)

    # -- probably not needed.
    # text = text.encode('unicode-escape').decode('ascii').replace('\\u', '\\')

    return text.replace("'", '\\27')


def style_for_print(options: Options) -> str:
    scss = f"""
    :root {{
        string-set: author '{_css_escape(options.author)}',
            copyright '{_css_escape(options.copyright)}',
            title '{_css_escape(options.cover_title)}';
    }}
    h1, h2, h3 {{
        string-set: chapter content();
    }}
    """
    css = sass.compile(string=scss)

    output_style = 'compressed'
    if options.debug_html:
        output_style = 'nested'

    base_path = os.path.abspath(os.path.dirname(__file__))

    filename = os.path.join(base_path, "report-print.scss")
    css += sass.compile(filename=filename, output_style=output_style)

    if options.cover or options.back_cover:
        filename = os.path.join(base_path, "cover.scss")
        css += sass.compile(filename=filename, output_style=output_style)

    filename = os.path.join(options.custom_template_path, 'styles.scss')
    if os.path.exists(filename):
        css += sass.compile(filename=filename, output_style=output_style)

    return css



from mkdocs.config.base import Config


class _FilterBase:
    def __init__(self, options: object, config: Config):
        self.__options = options
        self.__config = config

    @property
    def options(self):
        return self.__options

    @property
    def config(self):
        return self.__config

    def __call__(self, *args):
        raise "must be overridden"



import base64
import io

import barcode

from . import _FilterBase


#
# @see https://pypi.org/project/python-barcode/
#

class Barcode(_FilterBase):
    """ Generate Barcode for given value and returns embedded image data. """

    # Support:
    # ['code39', 'code128', 'ean', 'ean13', 'ean8',
    #  'gs1', 'gtin', 'isbn',
    #  'isbn10', 'isbn13', 'issn', 'jan',
    #  'pzn', 'upc', 'upca']

    def __call__(self, value, kind, **kwargs):
        coder = barcode.get_barcode_class(kind)
        img = coder(value, writer=None, **kwargs)
        with io.BytesIO() as stream:
            img.write(stream, dict(compress=True))
            data = base64.b64encode(stream.getvalue())
            return 'data:image/svg+xml;charset=utf-8;base64,' + \
                data.decode("ascii")



import os
from datetime import datetime

import jinja2
from mkdocs.config.base import Config

from .filters.datetime import strftime, strptime
from .filters.url import URLFilter

try:
    from .filters.barcode import Barcode
except ModuleNotFoundError:
    Barcode = None
try:
    from .filters.qrcode import QRCode
except ModuleNotFoundError:
    QRCode = None


class Template(object):

    """ Pickups key-value from `mkdocs.yml` """
    __KEYS = [
        "author", "copyright",
        "cover_title", "cover_subtitle", "cover_logo",

        "site_url",
        "repo_url"
    ]

    def __init__(self, options: object, config: Config):
        self._options = options
        self._config = config

        self._keywords = None
        self._jinja_env = None

    @property
    def _env(self) -> jinja2.Environment:

        def generate():
            base_path = os.path.abspath(os.path.dirname(__file__))
            template_paths = [os.path.join(base_path, '.')]

            if os.path.exists(self._options.custom_template_path):
                template_paths.append(self._options.custom_template_path)

            file_loader = jinja2.FileSystemLoader(template_paths)
            logging_undefined = jinja2.make_logging_undefined(
                logger=self._options.logger, base=jinja2.Undefined)
            env = jinja2.Environment(
                loader=file_loader,
                undefined=logging_undefined,
                lstrip_blocks=True,
                trim_blocks=True)

            env.filters['strptime'] = strptime
            env.filters['strftime'] = strftime

            env.filters['to_url'] = URLFilter(self._options, self._config)

            if Barcode:
                env.filters['barcode'] = Barcode(self._options, self._config)
            if QRCode:
                env.filters['qrcode'] = QRCode(self._options, self._config)

            return env

        if not self._jinja_env:
            self._jinja_env = generate()
        return self._jinja_env

    @property
    def keywords(self) -> dict:
        """ Keywords to pass when rendering the template. """

        import html

        def unescape_html_in_list(values: list) -> list:
            new_values = []
            for v in values:
                if isinstance(v, str):
                    new_values.append(html.unescape(v))
                elif isinstance(v, list):
                    new_values.append(unescape_html_in_list(v))
                elif isinstance(v, dict):
                    unescape_html(v)
                    new_values.append(v)
                else:
                    new_values.append(v)
            return new_values

        def unescape_html(variables: dict):
            for k, v in variables.items():
                if isinstance(v, str):
                    variables[k] = html.unescape(v)
                elif isinstance(v, list):
                    variables[k] = unescape_html_in_list(v)
                elif isinstance(v, dict):
                    unescape_html(v)

        def build_keywords():
            # keywords = {}
            keywords = self._config['extra']

            for key in self.__KEYS:
                if hasattr(self._options, key):
                    keywords[key] = getattr(self._options, key)
                elif key in self._config:
                    keywords[key] = self._config[key]

            unescape_html(keywords)

            keywords['now'] = datetime.now()

            if self._options.verbose or self._options.debug_html:
                from pprint import pformat
                self._options.logger.info('Template variables:')
                for line in pformat(keywords).split('\n'):
                    self._options.logger.info('  ' + line)

            return keywords

        if not self._keywords:
            self._keywords = build_keywords()

        return self._keywords

    def select(
        self, names: [str], parent=None, globals=None
    ) -> jinja2.Template:
        """ Find and load a template by names of given. """

        real_names = []
        for name in names:
            for ext in ['.html.j2', '.html.jinja2', '.html', '.htm']:
                real_names.append(name + ext)

        return self._env.select_template(
            real_names, parent=parent, globals=globals)
